"""Accessor class for columns containing single-level key/value mappings

The FlatSampleReader container is used to store data (in any backend) in a column
containing a single level key/value mapping from names/ids to data.

All backends are supported.
"""
from contextlib import ExitStack
from pathlib import Path
from typing import Tuple, Union, Iterable, Optional, Any

from .common import open_file_handles
from ..records import (
    data_record_db_val_from_digest,
    data_record_digest_val_from_db_val,
    flat_data_db_key_from_names,
    hash_data_db_key_from_raw_key,
    schema_db_key_from_column,
    schema_hash_db_key_from_digest,
    schema_hash_record_db_val_from_spec,
    schema_record_db_val_from_digest
)
from ..records.parsing import generate_sample_name
from ..backends import backend_decoder
from ..op_state import reader_checkout_only
from ..utils import is_suitable_user_key, valfilter, valfilterfalse


KeyType = Union[str, int]


class FlatSampleReader:
    """Class implementing get access to data in a column.

    This class exposes the standard API to access data stored in a single level
    key / value mapping column. Usage is modeled after the python :class:`dict`
    style syntax -- with a few additional utility and inspection methods and
    properties added. Methods named after those of a python :class:`dict` have
    syntactically identical arguments and behavior to that of the standard
    library.

    If not opened in a ``write-enabled`` checkout, then attempts to add or
    delete data or container properties will raise an exception (in the form of
    a :class:`PermissionError`). No changes will be propogated unless a
    ``write-enabled`` checkout is used.

    This object can be serialized -- pickled -- for parallel processing /
    reading if opened in a ``read-only`` checkout. Parallel operations are both
    thread and process safe, though performance may significantly differ
    between multithreaded vs multiprocessed code (depending on the backend data
    is stored in). Attempts to serialize objects opened in ``write-enabled``
    checkouts are not supported and will raise a :class:`PermissionError` if
    attempted. This behavior is enforced in order to ensure data and record
    integrity while writing to the repository.
    """

    __slots__ = ('_mode', '_column_name', '_samples', '_be_fs',
                 '_path', '_stack', '_enter_count', '_schema')
    _attrs = __slots__

    def __init__(self,
                 columnname: str,
                 samples,
                 backend_handles,
                 schema,
                 repo_path: Path,
                 mode: str,
                 *args, **kwargs):

        self._stack: Optional[ExitStack] = None
        self._mode = mode
        self._column_name = columnname
        self._samples = samples
        self._be_fs = backend_handles
        self._path = repo_path
        self._schema = schema
        self._enter_count = 0

    @property
    def _debug_(self):  # pragma: no cover
        return {
            '__class__': self.__class__,
            '_mode': self._mode,
            '_column_name': self._column_name,
            '_be_fs': self._be_fs,
            '_path': self._path,
            '_contains_subsamples': self.contains_subsamples,
            '_stack': self._stack._exit_callbacks if self._stack else self._stack,
            '_enter_count': self._enter_count,
        }

    def __repr__(self):
        res = (
            f'{self.__class__.__qualname__}('
            f'repo_pth={self._path}, '
            f'aset_name={self._column_name}, '
            f"{[f'{key}={val}, ' for key, val in self._schema.schema.items()]}, "
            f'mode={self._mode})')
        return res

    def _repr_pretty_(self, p, cycle):
        res = f'Hangar {self.__class__.__qualname__} \
                \n    Column Name              : {self._column_name}\
                \n    Writeable                : {self.iswriteable}\
                \n    Column Type              : {self.column_type}\
                \n    Column Layout            : {self.column_layout}\
                \n    Schema Type              : {self.schema_type}\
                \n    DType                    : {self.dtype}\
                \n    Shape                    : {self.shape}\
                \n    Number of Samples        : {self.__len__()}\
                \n    Partial Remote Data Refs : {bool(self.contains_remote_references)}\n'
        p.text(res)

    def _ipython_key_completions_(self):  # pragma: no cover
        """Let ipython know that any key based access can use the column keys

        Since we don't want to inherit from dict, nor mess with `__dir__` for
        the sanity of developers, this is the best way to ensure users can
        autocomplete keys.

        Returns
        -------
        list
            list of strings, each being one of the column keys for access.
        """
        return list(self.keys())

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        return {slot: getattr(self, slot) for slot in self.__slots__}

    def __setstate__(self, state: dict) -> None:
        """ensure multiprocess operations can pickle relevant data.

        Technically should be decorated with @reader_checkout_only, but since
        at instance creation that is not an attribute, the decorator won't
        know. Since only readers can be pickled, This isn't much of an issue.
        """
        for slot, value in state.items():
            setattr(self, slot, value)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return

    def _destruct(self):
        if isinstance(self._stack, ExitStack):
            self._stack.close()
        self._close()
        for attr in self._attrs:
            delattr(self, attr)

    def __getattr__(self, name):
        """Raise permission error after checkout is closed.

         Only runs after a call to :meth:`_destruct`, which is responsible for
         deleting all attributes from the object instance.
        """
        try:
            self.__getattribute__('_mode')  # once checkout is closed, this won't exist.
        except AttributeError:
            err = (f'Unable to operate on past checkout objects which have been '
                   f'closed. No operation occurred. Please use a new checkout.')
            raise PermissionError(err) from None
        return self.__getattribute__(name)

    @property
    def _is_conman(self) -> bool:
        return bool(self._enter_count)

    def __iter__(self) -> Iterable[KeyType]:
        """Create iterator yielding an column sample keys.

        Yields
        -------
        Iterable[KeyType]
            Sample key contained in the column.
        """
        yield from self.keys()

    def __len__(self) -> int:
        """Check how many samples are present in a given column.
        """
        return len(self._samples)

    def __contains__(self, key: KeyType) -> bool:
        """Determine if a key is a valid sample name in the column.
        """
        return key in self._samples

    def _open(self):
        for val in self._be_fs.values():
            val.open(mode=self._mode)

    def _close(self):
        for val in self._be_fs.values():
            val.close()

    def __getitem__(self, key: KeyType):
        """Retrieve data for some sample key via dict style access conventions.

        .. seealso:: :meth:`get`

        Parameters
        ----------
        key : KeyType
            Sample key to retrieve from the column.

        Returns
        -------
        value
            Data corresponding to the provided sample key.

        Raises
        ------
        KeyError
            if no sample with the requested key exists.
        """
        spec = self._samples[key]
        return self._be_fs[spec.backend].read_data(spec)

    def get(self, key: KeyType, default=None):
        """Retrieve the data associated with some sample key

        Parameters
        ----------
        key : KeyType
            The name of the subsample(s) to retrieve. Passing a single
            subsample key will return the stored data value.
        default : Any
            if a `key` parameter is not found, then return this value instead.
            By default, None.

        Returns
        -------
        value
            data data stored under subsample key if key exists, else
            default value if not found.
        """
        try:
            return self[key]
        except KeyError:
            return default

    @property
    def column(self) -> str:
        """Name of the column.
        """
        return self._column_name

    @property
    def column_type(self):
        """Data container type of the column ('ndarray', 'str', etc).
        """
        return self._schema.column_type

    @property
    def column_layout(self):
        """Column layout type ('nested', 'flat', etc).
        """
        return self._schema.column_layout

    @property
    def schema_type(self):
        """Schema type of the contained data ('variable_shape', 'fixed_shape', etc).
        """
        return self._schema.schema_type

    @property
    def dtype(self):
        """Dtype of the columns data (np.float, str, etc).
        """
        return self._schema.dtype

    @property
    def shape(self):
        """(Max) shape of data that can (is) written in the column.
        """
        try:
            return self._schema.shape
        except AttributeError:
            return None

    @property
    def backend(self) -> str:
        """Code indicating which backing store is used when writing data.
        """
        return self._schema.backend

    @property
    def backend_options(self):
        """Filter / Compression options applied to backend when writing data.
        """
        return self._schema.backend_options

    @property
    def iswriteable(self) -> bool:
        """Bool indicating if this column object is write-enabled.
        """
        return False if self._mode == 'r' else True

    @property
    def contains_subsamples(self) -> bool:
        """Bool indicating if sub-samples are contained in this column container.
        """
        return False

    @property
    def contains_remote_references(self) -> bool:
        """Bool indicating if all samples in column exist on local disk.

        The data associated with samples referencing some remote server will
        need to be downloaded (``fetched`` in the hangar vocabulary) before
        they can be read into memory.

        Returns
        -------
        bool
            False if at least one sample in the column references data stored
            on some remote server. True if all sample data is available on the
            machine's local disk.
        """
        return not all(map(lambda x: x.islocal, self._samples.values()))

    @property
    def remote_reference_keys(self) -> Tuple[KeyType]:
        """Compute sample names whose data is stored in a remote server reference.

        Returns
        -------
        Tuple[KeyType]
            list of sample keys in the column whose data references indicate
            they are stored on a remote server.
        """
        return tuple(valfilterfalse(lambda x: x.islocal, self._samples).keys())

    def _mode_local_aware_key_looper(self, local: bool) -> Iterable[KeyType]:
        """Generate keys for iteration with dict update safety ensured.

        Parameters
        ----------
        local : bool
            True if keys should be returned which only exist on the local machine.
            Fale if remote sample keys should be excluded.

        Returns
        -------
        Iterable[KeyType]
            Sample keys conforming to the `local` argument spec.
        """
        if local:
            if self._mode == 'r':
                yield from valfilter(lambda x: x.islocal, self._samples).keys()
            else:
                yield from tuple(valfilter(lambda x: x.islocal, self._samples).keys())
        else:
            if self._mode == 'r':
                yield from self._samples.keys()
            else:
                yield from tuple(self._samples.keys())

    def keys(self, local: bool = False) -> Iterable[KeyType]:
        """Generator yielding the name (key) of every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys will only correspond to data which is
            available for reading on the local disk, by default False.

        Yields
        ------
        Iterable[KeyType]
            Keys of one subsample at a time inside the sample.
        """
        yield from self._mode_local_aware_key_looper(local)

    def values(self, local: bool = False) -> Iterable[Any]:
        """Generator yielding the data for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned values will only correspond to data which is
            available for reading on the local disk. No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Any]
            Values of one subsample at a time inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield self[key]

    def items(self, local: bool = False) -> Iterable[Tuple[KeyType, Any]]:
        """Generator yielding (name, data) tuple for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys/values will only correspond to data which is
            available for reading on the local disk, No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Tuple[KeyType, Any]]
            Name and stored value for every subsample inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield (key, self[key])

# ---------------- writer methods only after this point -------------------


class FlatSampleWriter(FlatSampleReader):

    __slots__ = ('_txnctx',)
    _attrs = __slots__ + FlatSampleReader.__slots__

    def __init__(self, aset_ctx, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._txnctx = aset_ctx

    def __enter__(self):
        with ExitStack() as stack:
            self._txnctx.open_write()
            stack.callback(self._txnctx.close_write)
            if self._enter_count == 0:
                for k in self._be_fs.keys():
                    stack.enter_context(self._be_fs[k])
            self._enter_count += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._stack.close()
        self._enter_count -= 1

    def _set_arg_validate(self, key, value):
        """Verify if key / value pair is valid to be written in this column

        Parameters
        ----------
        key
            name to associate with this data piece
        value
            piece of data to store in the column

        Raises
        ------
        ValueError
            If key is not valid type/contents or if value is not correct object
            type / if it does not conform to column schema
        """
        if not is_suitable_user_key(key):
            raise ValueError(f'Sample name `{key}` is not suitable.')

        isCompat = self._schema.verify_data_compatible(value)
        if not isCompat.compatible:
            raise ValueError(isCompat.reason)

    def _perform_set(self, key, value):
        """Internal write method. Assumes all arguments validated and context is open

        Parameters
        ----------
        key
            sample key to store
        value
            data to store
        """
        full_hash = self._schema.data_hash_digest(value)

        hashKey = hash_data_db_key_from_raw_key(full_hash)
        # check if data record already exists with given key
        dataRecKey = flat_data_db_key_from_names(self._column_name, key)
        existingDataRecVal = self._txnctx.dataTxn.get(dataRecKey, default=False)
        if existingDataRecVal:
            # check if data record already with same key & hash value
            existingDataRec = data_record_digest_val_from_db_val(existingDataRecVal)
            if full_hash == existingDataRec.digest:
                return

        # write new data if data hash does not exist
        existingHashVal = self._txnctx.hashTxn.get(hashKey, default=False)
        if existingHashVal is False:
            hashVal = self._be_fs[self._schema.backend].write_data(value)
            self._txnctx.hashTxn.put(hashKey, hashVal)
            self._txnctx.stageHashTxn.put(hashKey, hashVal)
            hash_spec = backend_decoder(hashVal)
        else:
            hash_spec = backend_decoder(existingHashVal)
            if hash_spec.backend not in self._be_fs:
                # when adding data which is already stored in the repository, the
                # backing store for the existing data location spec may not be the
                # same as the backend which the data piece would have been saved in here.
                #
                # As only the backends actually referenced by a columns samples are
                # initialized (accessible by the column), there is no guarantee that
                # an accessor exists for such a sample. In order to prevent internal
                # errors from occurring due to an uninitialized backend if a previously
                # existing data piece is "saved" here and subsequently read back from
                # the same writer checkout, we perform an existence check and backend
                # initialization, if appropriate.
                fh = open_file_handles(backends=(hash_spec.backend,),
                                       path=self._path,
                                       mode='a',
                                       schema=self._schema)
                self._be_fs[hash_spec.backend] = fh[hash_spec.backend]

        # add the record to the db
        dataRecVal = data_record_db_val_from_digest(full_hash)
        self._txnctx.dataTxn.put(dataRecKey, dataRecVal)
        self._samples[key] = hash_spec

    def __setitem__(self, key, value):
        """Store a piece of data in a column.

        .. seealso::

            :meth:`update` for an implementation analogous to python's built in
            :meth:`dict.update` method which accepts a dict or iterable of
            key/value pairs to add in the same operation.

        Parameters
        ----------
        key
            name to assign to the sample (assuming the column accepts named
            samples), If str, can only contain alpha-numeric ascii characters
            (in addition to '-', '.', '_'). Integer key must be >= 0. by
            default, None
        value
            data to store as a sample in the column.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)
            self._set_arg_validate(key, value)
            self._perform_set(key, value)

    def append(self, value) -> KeyType:
        """Store some data in a sample with an automatically generated key.

        This method should only be used if the context some piece of data is
        used in is independent from it's value (ie. when reading data back,
        there is no useful information which needs to be conveyed between the
        data source's name/id and the value of that piece of information.)
        Think carefully before going this route, as this posit does not apply
        to many common use cases.

        To store the data with a user defined key, use :meth:`update` or
        :meth:`__setitem__`

        Parameters
        ----------
        value
            Piece of data to store in the column.

        Returns
        -------
        KeyType
            Name of the generated key this data is stored with.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)
            key = generate_sample_name()
            while key in self._samples:
                key = generate_sample_name()
            self._set_arg_validate(key, value)
            self._perform_set(key, value)
            return key

    def update(self, other=None, **kwargs):
        """Store some data with the key/value pairs from other, overwriting existing keys.

        :meth:`update` implements functionality similar to python's builtin
        :meth:`dict.update` method, accepting either a dictionary or other
        iterable (of length two) listing key / value pairs.

        Parameters
        ----------
        other
            Accepts either another dictionary object or an iterable of
            key/value pairs (as tuples or other iterables of length two).
            mapping sample names to data value instances instances, If sample
            name is string type, can only contain alpha-numeric ascii
            characters (in addition to '-', '.', '_'). Int key must be >= 0. By
            default, None.
        **kwargs
            keyword arguments provided will be saved with keywords as sample keys
            (string type only) and values as np.array instances.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if other:
                if not isinstance(other, dict):
                    other = dict(other)
                else:
                    other = other.copy()
            elif other is None:
                other = {}
            if kwargs:
                # we have to merge kwargs dict with `other` before operating on
                # either so all validation and writing occur atomically
                other.update(kwargs)

            for key, val in other.items():
                self._set_arg_validate(key, val)
            for key, val in other.items():
                self._perform_set(key, val)

    def __delitem__(self, key: KeyType) -> None:
        """Remove a sample from the column. Convenience method to :meth:`delete`.

        .. seealso::

            :meth:`pop` to return a value and then delete it in the same operation

        Parameters
        ----------
        key : KeyType
            Name of the sample to remove from the column.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if key not in self._samples:
                raise KeyError(key)

            dataKey = flat_data_db_key_from_names(self._column_name, key)
            isRecordDeleted = self._txnctx.dataTxn.delete(dataKey)
            if isRecordDeleted is False:
                raise RuntimeError(
                    f'Internal error. Not able to delete key {key} from staging '
                    f'db even though existance passed in memory verification. '
                    f'Please report this message in full to the hangar development team.',
                    f'Specified key: <{type(key)} {key}>', f'Calculated dataKey: <{dataKey}>',
                    f'isRecordDeleted: <{isRecordDeleted}>', f'DEBUG STRING: {self._debug_}')
            del self._samples[key]

    def pop(self, key: KeyType):
        """Retrieve some value for some key(s) and delete it in the same operation.

        Parameters
        ----------
        key : KeysType
            Sample key to remove

        Returns
        -------
        value
            Upon success, the value of the removed key.

        Raises
        ------
        KeyError
            If there is no sample with some key in the column.
        """
        value = self[key]
        del self[key]
        return value

    def change_backend(self, backend: str, backend_options: Optional[dict] = None):
        """Change the default backend and filters applied to future data writes.

        .. warning::

           This method is meant for advanced users only. Please refer to the
           hangar backend codebase for information on accepted parameters and
           options.

        Parameters
        ----------
        backend : str
            Backend format code to swtich to.
        backend_options : Optional[dict]
            Backend option specification to use (if specified). If left to
            default value of None, then default options for backend are
            automatically used.

        Raises
        ------
        RuntimeError
            If this method was called while this column is invoked in a
            context manager
        ValueError
            If the backend format code is not valid.
        """
        if self._is_conman:
            raise RuntimeError('Cannot call method inside column context manager.')

        self._schema.change_backend(backend, backend_options=backend_options)

        new_schema_digest = self._schema.schema_hash_digest()
        columnSchemaKey = schema_db_key_from_column(self._column_name, layout=self.column_layout)
        columnSchemaVal = schema_record_db_val_from_digest(new_schema_digest)
        hashSchemaKey = schema_hash_db_key_from_digest(new_schema_digest)
        hashSchemaVal = schema_hash_record_db_val_from_spec(self._schema.schema)

        # -------- set vals in lmdb only after schema is sure to exist --------

        with self._txnctx.write() as ctx:
            ctx.dataTxn.put(columnSchemaKey, columnSchemaVal)
            ctx.hashTxn.put(hashSchemaKey, hashSchemaVal, overwrite=False)

        new_backend = self._schema.backend
        if new_backend not in self._be_fs:
            fhands = open_file_handles(
                backends=[new_backend],
                path=self._path,
                mode='a',
                schema=self._schema)
            self._be_fs[new_backend] = fhands[new_backend]
        else:
            self._be_fs[new_backend].close()
        self._be_fs[new_backend].open(mode='a')
        self._be_fs[new_backend].backend_opts = self._schema.backend_options
        return



"""Local HDF5 Backend Implementation, Identifier: ``HDF5_01``

Backend Identifiers
===================

*  Backend: ``0``
*  Version: ``1``
*  Format Code: ``01``
*  Canonical Name: ``HDF5_01``

Storage Method
==============

*  This module is meant to handle larger datasets which are of fixed size. IO
   and significant compression optimization is achieved by storing arrays at
   their appropriate top level index in the same shape they naturally assume
   and chunking over the entire subarray domain making up a sample (rather than
   having to subdivide chunks when the sample could be variably shaped.)

*  Data is written to specific subarray indexes inside an HDF5 "dataset" in a
   single HDF5 File.

*  In each HDF5 File there are ``COLLECTION_COUNT`` "datasets" (named ``["0" :
   "{COLLECTION_COUNT}"]``). These are referred to as ``"dataset number"``

*  Each dataset is a zero-initialized array of:

   *  ``dtype: {schema_dtype}``; ie ``np.float32`` or ``np.uint8``

   *  ``shape: (COLLECTION_SIZE, *{schema_shape})``; ie ``(500, 10, 10)`` or
      ``(500, 512, 512, 320)``. The first index in the dataset is referred to as a
      ``collection index``.

*  Compression Filters, Chunking Configuration/Options are applied globally for
   all ``datasets`` in a file at dataset creation time.

*  On read and write of all samples the xxhash64_hexdigest is calculated for
   the raw array bytes. This is to ensure that all data in == data out of the
   hdf5 files. That way even if a file is manually edited (bypassing fletcher32
   filter check) we have a quick way to tell that things are not as they should
   be.

Compression Options
===================

Accepts dictionary containing keys

*  ``backend`` == ``"01"``
*  ``complib``
*  ``complevel``
*  ``shuffle``

Blosc-HDF5

*  ``complib`` valid values:

   *  ``'blosc:blosclz'``,
   *  ``'blosc:lz4'``,
   *  ``'blosc:lz4hc'``,
   *  ``'blosc:zlib'``,
   *  ``'blosc:zstd'``

*  ``complevel`` valid values: [0, 9] where 0 is "no compression" and 9 is
   "most compression"

*  ``shuffle`` valid values:

   *  ``None``
   *  ``'none'``
   *  ``'byte'``
   *  ``'bit'``


LZF Filter

*  ``'complib' == 'lzf'``
*  ``'shuffle'`` one of ``[False, None, 'none', True, 'byte']``
*  ``'complevel'`` one of ``[False, None, 'none']``

GZip Filter

*  ``'complib' == 'gzip'``
*  ``'shuffle'`` one of ``[False, None, 'none', True, 'byte']``
*  ``complevel`` valid values: [0, 9] where 0 is "no compression" and 9 is
   "most compression"


Record Format
=============

Fields Recorded for Each Array
------------------------------

*  Format Code
*  File UID
*  xxhash64_hexdigest (ie. checksum)
*  Dataset Number (``0:COLLECTION_COUNT`` dataset selection)
*  Dataset Index (``0:COLLECTION_SIZE`` dataset subarray selection)
*  Subarray Shape

Examples
--------

1)  Adding the first piece of data to a file:

    *  Array shape (Subarray Shape): (10, 10)
    *  File UID: "rlUK3C"
    *  xxhash64_hexdigest: 8067007c0f05c359
    *  Dataset Number: 16
    *  Collection Index: 105

    ``Record Data => "01:rlUK3C:8067007c0f05c359:16:105:10 10"``

1)  Adding to a piece of data to a the middle of a file:

    *  Array shape (Subarray Shape): (20, 2, 3)
    *  File UID: "rlUK3C"
    *  xxhash64_hexdigest: b89f873d3d153a9c
    *  Dataset Number: "3"
    *  Collection Index: 199

    ``Record Data => "01:rlUK3C:b89f873d3d153a9c:8:199:20 2 3"``


Technical Notes
===============

*  The majority of methods not directly related to "chunking" and the "raw data
   chunk cache" are either identical to HDF5_00, or only slightly modified.

*  Files are read only after initial creation/writes. Only a write-enabled
   checkout can open a HDF5 file in ``"w"`` or ``"a"`` mode, and writer
   checkouts create new files on every checkout, and make no attempt to fill in
   unset locations in previous files. This is not an issue as no disk space is
   used until data is written to the initially created "zero-initialized"
   collection datasets

*  On write: Single Writer Multiple Reader (``SWMR``) mode is set to ensure that
   improper closing (not calling ``.close()``) method does not corrupt any data
   which had been previously flushed to the file.

*  On read: SWMR is set to allow multiple readers (in different threads /
   processes) to read from the same file. File handle serialization is handled
   via custom python ``pickle`` serialization/reduction logic which is
   implemented by the high level ``pickle`` reduction ``__set_state__()``,
   ``__get_state__()`` class methods.

*  An optimization is performed in order to increase the read / write
   performance of fixed size datasets. Due to the way that we initialize an
   entire HDF5 file with all datasets pre-created (to the size of the fixed
   subarray shape), and the fact we absolutely know the size / shape /
   access-pattern of the arrays, inefficient IO due to wasted chunk processing
   is not a concern. It is far more efficient for us to completely blow off the
   metadata chunk cache, and chunk each subarray as a single large item item.

   This method of processing tends to have a number of significant effects as
   compared to chunked storage methods:

      1. **Compression rations improve** (by a non-trivial factor). This is
         simply due to the fact that a larger amount of raw data is being passed
         into the compressor at a time. While the exact improvement seen is
         highly dependent on both the data size and compressor used, there
         should be no case where compressing the full tensor uses more disk
         space then chunking the tensor, compressing each chunk individually,
         and then saving each chunk to disk.

      2. **Read performance improves** (so long as a suitable compressor /
         option set was chosen). Instead of issuing (potentially) many read
         requests - one for each chunk - to the storage hardware, signifiantly
         few IOPS are used to retrieve the entire set of compressed raw data
         from disk. Fewer IOPS means much less time waiting on the hard disk.
         Moreso, only a single decompression step is needed to reconstruct
         the numeric array, completly decoupling performance from HDF5's
         ability to parallelize internal filter pipeline operations.

         Additionally, since the entire requested chunk is retrieved in a
         single decompression pipeline run, there is no need for the HDF5 core
         to initialize an intermediate buffer which holds data chunks as each
         decompression operation completes. Futher, by preinitializing an empty
         ``numpy.ndarray`` container and using the low level HDF5
         ``read_direct`` method, the decompressed data buffer is passes
         directly into the returned ``ndarray.__array_interface__.data``
         field with no intermediate copy or processing steps.

      3. **Shuffle filters are favored.**. With much more data to work with in
         a single compression operation, the use of "byte shuffle" filters in
         the compressor spec has been seen to both markedly decrease read time
         and increase compression ratios. Shuffling can significantly reduce
         disk space required to store some piece of data on disk, further
         reducing the time spent waiting on hard disk IO while incuring a
         negligible cost to decompression speed.

   Taking all of these effects into account, there can be up to an order of
   magnitude increase in read performance as compared to the subarray chunking
   strategy employed by the ``HDF5_00`` backend.

*  Like all other backends at the time of writing, only 'C' ordered arrays
   are accepted by this method.
"""
import logging
import math
import os
from collections import ChainMap
from contextlib import suppress
from functools import partial
from pathlib import Path
from typing import MutableMapping, Tuple, Optional, Union, Callable

import h5py
import numpy as np

try:
    # hdf5plugin warns if a filter is already loaded.
    _logger = logging.getLogger('hdf5plugin')
    _initialLevel = _logger.getEffectiveLevel()
    _logger.setLevel(logging.ERROR)
    import hdf5plugin
    if 'blosc' not in hdf5plugin.FILTERS:
        raise ImportError(f'BLOSC unavailable via hdf5plugin: {hdf5plugin.FILTERS}')
finally:
    _logger.setLevel(_initialLevel)
from xxhash import xxh64_hexdigest


from .chunk import calc_chunkshape
from .specs import HDF5_01_DataHashSpec
from .. import __version__
from ..optimized_utils import SizedDict
from ..constants import DIR_DATA_REMOTE, DIR_DATA_STAGE, DIR_DATA_STORE, DIR_DATA
from ..op_state import writer_checkout_only, reader_checkout_only
from ..utils import find_next_prime, random_string, set_blosc_nthreads
from ..typesystem import Descriptor, OneOf, DictItems, SizedIntegerTuple, checkedmeta

set_blosc_nthreads()

# ----------------------------- Configuration ---------------------------------

_FmtCode = '01'

# contents of a single hdf5 file
COLLECTION_SIZE = 100
COLLECTION_COUNT = 100
CHUNK_MAX_RDCC_NBYTES = 250_000_000
CHUNK_RDCC_W0 = 0.75

# -------------------------------- Parser Implementation ----------------------


def hdf5_01_encode(uid: str, cksum: str, dset: int, dset_idx: int,
                   shape: Tuple[int]) -> bytes:
    """converts the hdf5 data has spec to an appropriate db value

    Parameters
    ----------
    uid : str
        the file name prefix which the data is written to.
    cksum : int
        xxhash_64.hex_digest checksum of the data bytes in numpy array form.
    dset : int
        collection (ie. hdf5 dataset) name to find this data piece.
    dset_idx : int
        collection first axis index in which this data piece resides.
    shape : Tuple[int]
        shape of the data sample written to the collection idx. ie:
        what subslices of the hdf5 dataset should be read to retrieve
        the sample as recorded.

    Returns
    -------
    bytes
        hash data db value recording all input specifications.
    """
    shape_str = " ".join([str(i) for i in shape])
    return f'01:{uid}:{cksum}:{dset}:{dset_idx}:{shape_str}'.encode()


# ------------------------- Accessor Object -----------------------------------


@DictItems(
    expected_keys_required={'complib': True, 'complevel': True, 'shuffle': True},
    expected_values={
        'complib': ['blosc:blosclz', 'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd'],
        'complevel': [i for i in range(10)],
        'shuffle': [None, 'none', 'byte', 'bit']})
class BloscCompressionOptions(Descriptor):
    pass


@DictItems(
    expected_keys_required={'complib': True, 'complevel': True, 'shuffle': True},
    expected_values={
        'complib': ['gzip'], 'complevel': [i for i in range(10)], 'shuffle': [True, False]})
class GzipCompressionOptions(Descriptor):
    pass


@DictItems(
    expected_keys_required={'complib': True, 'complevel': False, 'shuffle': True},
    expected_values={
        'complib': ['lzf'], 'complevel': ['none', None], 'shuffle': [True, False]})
class LzfCompressionOptions(Descriptor):
    pass


@OneOf(list(map(lambda x: np.dtype(x).name, [
        np.bool, np.uint8, np.uint16, np.uint32, np.uint64, np.int8, np.int16,
        np.int32, np.int64, np.float16, np.float32, np.float64, np.longdouble])))
class AllowedDtypes(Descriptor):
    # Note. np.longdouble since np.float128 not guaranteed to be available on
    # all system. this is a particular issue with some windows numpy builds.
    pass


class HDF5_01_Options(metaclass=checkedmeta):
    _shape = SizedIntegerTuple(size=32)
    _dtype = AllowedDtypes()
    _lzf = LzfCompressionOptions()
    _gzip = GzipCompressionOptions()
    _blosc = BloscCompressionOptions()
    _avail_filters = ('_lzf', '_gzip', '_blosc')

    def __init__(self, backend_options, dtype, shape, *args, **kwargs):
        self._shape = shape
        self._dtype = dtype
        self._selected_filter = None
        if backend_options is None:
            backend_options = self.default_options

        for filter_attr in self._avail_filters:
            with suppress((KeyError, ValueError)):
                setattr(self, filter_attr, backend_options)
                self._selected_filter = filter_attr
                break
        else:  # N.B. for-else loop (ie. "no-break")
            raise ValueError(f'Invalid backend_options {backend_options}')
        self._verify_data_nbytes_larger_than_clib_min()

    def _verify_data_nbytes_larger_than_clib_min(self):
        """blosc clib should not be used if data buffer size < 16 bytes.

        Raises
        ------
        ValueError:
            if the data size is not valid for the clib
        """
        if self._selected_filter in ['_blosc', None]:
            num_items = np.prod(self._shape)
            itemsize = np.dtype(self._dtype).itemsize
            nbytes = itemsize * num_items
            if nbytes <= 16:
                raise ValueError(f'blosc clib requires data buffer size > 16 bytes')

    @property
    def default_options(self):
        if 'blosc' in hdf5plugin.FILTERS:
            try:
                self._verify_data_nbytes_larger_than_clib_min()
                return {'complib': 'blosc:lz4hc', 'complevel': 5, 'shuffle': 'byte'}
            except ValueError:
                pass
        return {'complib': 'lzf', 'complevel': None, 'shuffle': True}

    @property
    def backend_options(self):
        return getattr(self, self._selected_filter)

    @property
    def init_requires(self):
        return ('repo_path', 'schema_shape', 'schema_dtype')


HDF5_01_MapTypes = MutableMapping[str, Union[h5py.File, Callable[[], h5py.File]]]


class HDF5_01_FileHandles(object):
    """Manage HDF5 file handles.

    When in SWMR-write mode, no more than a single file handle can be in the
    "writeable" state. This is an issue where multiple columns may need to
    write to the same column schema.
    """

    def __init__(self, repo_path: Path, schema_shape: tuple, schema_dtype: np.dtype):
        self.path: Path = repo_path
        self.schema_shape: tuple = schema_shape
        self.schema_dtype: np.dtype = schema_dtype
        self._dflt_backend_opts: Optional[dict] = None

        self.rFp: HDF5_01_MapTypes = {}
        self.wFp: HDF5_01_MapTypes = {}
        self.Fp: HDF5_01_MapTypes = ChainMap(self.rFp, self.wFp)
        self.rDatasets = SizedDict(maxsize=100)
        self.wdset: h5py.Dataset = None

        self.mode: Optional[str] = None
        self.hIdx: Optional[int] = None
        self.w_uid: Optional[str] = None
        self.hMaxSize: Optional[int] = None
        self.hNextPath: Optional[int] = None
        self.hColsRemain: Optional[int] = None

        self.STAGEDIR: Path = Path(self.path, DIR_DATA_STAGE, _FmtCode)
        self.REMOTEDIR: Path = Path(self.path, DIR_DATA_REMOTE, _FmtCode)
        self.DATADIR: Path = Path(self.path, DIR_DATA, _FmtCode)
        self.STOREDIR: Path = Path(self.path, DIR_DATA_STORE, _FmtCode)
        self.DATADIR.mkdir(exist_ok=True)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        if self.w_uid in self.wFp:
            self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
            self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
            self.wFp[self.w_uid].flush()

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        self.close()
        state = self.__dict__.copy()
        del state['rFp']
        del state['wFp']
        del state['Fp']
        del state['rDatasets']
        del state['wdset']
        return state

    def __setstate__(self, state: dict) -> None:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.__dict__.update(state)
        self.rFp = {}
        self.wFp = {}
        self.Fp = ChainMap(self.rFp, self.wFp)
        self.rDatasets = {}
        self.wdset = None
        self.open(mode=self.mode)

    @property
    def backend_opts(self):
        return self._dflt_backend_opts

    @writer_checkout_only
    def _backend_opts_set(self, val):
        """Nonstandard descriptor method. See notes in ``backend_opts.setter``.
        """
        self._dflt_backend_opts = val
        return

    @backend_opts.setter
    def backend_opts(self, value):
        """
        Using seperate setter method (with ``@writer_checkout_only`` decorator
        applied) due to bug in python <3.8.

        From: https://bugs.python.org/issue19072
            > The classmethod decorator when applied to a function of a class,
            > does not honour the descriptor binding protocol for whatever it
            > wraps. This means it will fail when applied around a function which
            > has a decorator already applied to it and where that decorator
            > expects that the descriptor binding protocol is executed in order
            > to properly bind the function to the class.
        """
        return self._backend_opts_set(value)

    def open(self, mode: str, *, remote_operation: bool = False):
        """Open an hdf5 file handle in the Handler Singleton

        Parameters
        ----------
        mode : str
            one of `r` or `a` for read only / read-write.
        remote_operation : optional, kwarg only, bool
            if this hdf5 data is being created from a remote fetch operation, then
            we don't open any files for reading, and only open files for writing
            which exist in the remote data dir. (default is false, which means that
            write operations use the stage data dir and read operations use data store
            dir)
        """
        self.mode = mode
        if self.mode == 'a':
            process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
            process_dir.mkdir(exist_ok=True)
            for uidpth in process_dir.iterdir():
                if uidpth.suffix == '.hdf5':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(
                        h5py.File, file_pth, 'r', swmr=True, libver='latest')

        if not remote_operation:
            if not self.STOREDIR.is_dir():
                return
            for uidpth in self.STOREDIR.iterdir():
                if uidpth.suffix == '.hdf5':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(
                        h5py.File, file_pth, 'r', swmr=True, libver='latest')

    def close(self):
        """Close a file handle after writes have been completed

        behavior changes depending on write-enable or read-only file

        Returns
        -------
        bool
            True if success, otherwise False.
        """
        if self.mode == 'a':
            if self.w_uid in self.wFp:
                self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
                self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
                self.wFp[self.w_uid].flush()
            for uid in list(self.wFp.keys()):
                with suppress(AttributeError):
                    self.wFp[uid].close()
                del self.wFp[uid]
            self.wdset = None
            self.hMaxSize = None
            self.hNextPath = None
            self.hIdx = None
            self.hColsRemain = None
            self.w_uid = None

        for uid in list(self.rFp.keys()):
            with suppress(AttributeError):
                self.rFp[uid].close()
            del self.rFp[uid]
        self.rDatasets = {}

    @staticmethod
    def delete_in_process_data(repo_path: Path, *, remote_operation=False) -> None:
        """Removes some set of files entirely from the stage/remote directory.

        DANGER ZONE. This should essentially only be used to perform hard resets
        of the repository state.

        Parameters
        ----------
        repo_path : Path
            path to the repository on disk
        remote_operation : optional, kwarg only, bool
            If true, modify contents of the remote_dir, if false (default) modify
            contents of the staging directory.
        """
        data_dir = Path(repo_path, DIR_DATA, _FmtCode)
        PDIR = DIR_DATA_STAGE if not remote_operation else DIR_DATA_REMOTE
        process_dir = Path(repo_path, PDIR, _FmtCode)
        if not process_dir.is_dir():
            return

        for uidpth in process_dir.iterdir():
            if uidpth.suffix == '.hdf5':
                os.remove(process_dir.joinpath(uidpth.name))
                os.remove(data_dir.joinpath(uidpth.name))
        os.rmdir(process_dir)

    @staticmethod
    def _dataset_opts(complib: str, complevel: int, shuffle: Union[bool, str]) -> dict:
        """specify compression options for the hdf5 dataset.

        .. seealso:: :function:`_blosc_opts`

        to enable blosc compression, use the conda-forge `blosc-hdf5-plugin` package.

        .. seealso::

        * https://github.com/conda-forge/staged-recipes/pull/7650
        * https://github.com/h5py/h5py/issues/611

        Parameters
        ----------
        complib : str
            the compression lib to use, one of ['lzf', 'gzip', 'blosc:blosclz',
            'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd']
        complevel : int
            compression level to specify (accepts values [0, 9] for all except 'lzf'
            where no complevel is accepted)
        shuffle : bool
            if True or `byte`, enable byte shuffle filter, if blosc
            compression, pass through 'bits' is accepted as well. False, or
            None indicates no shuffle should be applied.
        """
        # ---- blosc hdf5 plugin filters ----
        _blosc_compression = {
            'blosc:blosclz': 0,
            'blosc:lz4': 1,
            'blosc:lz4hc': 2,
            # Not built 'snappy': 3,
            'blosc:zlib': 4,
            'blosc:zstd': 5}
        _blosc_shuffle = {None: 0, 'none': 0, 'byte': 1, 'bit': 2}
        _blosc_complevel = {**{i: i for i in range(10)}, None: 9, 'none': 9}

        # ---- h5py built in filters ----
        _lzf_gzip_shuffle = {None: False, False: False, 'none': False, True: True, 'byte': True}
        _lzf_complevel = {False: None, None: None, 'none': None}
        _gzip_complevel = {**{i: i for i in range(10)}, None: 4, 'none': 4}

        if complib.startswith('blosc'):
            args = {
                'compression': 32001,
                'compression_opts': (
                    0, 0, 0, 0,
                    _blosc_complevel[complevel],
                    _blosc_shuffle[shuffle],
                    _blosc_compression[complib]),
                'shuffle': False}
        elif complib == 'lzf':
            args = {
                'shuffle': _lzf_gzip_shuffle[shuffle],
                'compression': complib,
                'compression_opts': _lzf_complevel[complevel]}
        elif complib == 'gzip':
            args = {
                'shuffle': _lzf_gzip_shuffle[shuffle],
                'compression': complib,
                'compression_opts': _gzip_complevel[complevel]}
        elif complib in (None, False, 'none'):
            args = {
                'shuffle': False,
                'compression': None,
                'compression_opts': None}
        else:
            raise ValueError(f'unknown value for opt arg `complib`: {complib}')
        return args

    def _create_schema(self, *, remote_operation: bool = False):
        """stores the shape and dtype as the schema of a column.

        Parameters
        ----------
        remote_operation : optional, kwarg only, bool
            if this schema is being created from a remote fetch operation, then do not
            place the file symlink in the staging directory. Instead symlink it
            to a special remote staging directory. (default is False, which places the
            symlink in the stage data directory.)

        Notes
        -----

        Parameters set for raw-data-chunk-cache (rdcc) values:

        * rdcc_nbytes: sets the total size (measured in bytes) of the raw data chunk
          cache for each dataset. This should be set to the size of each chunk times
          the number of chunks that are likely to be needed in cache.
        * rdcc_w0: sets the policy for chunks to be removed from the cache when more
          space is needed. If set to 0, always evict the least recently used chunk in
          cache. If set to 1, always evict the least recently used chunk which has
          been fully read or written. If the value is between 0 and 1, the behavior
          will be a blend of the two.
        * rdcc_nslots: The number of chunk slots in the cache for this entire file.
          In order for quick lookup, a hash map is used for each chunk value. For
          maximum performance, this value should be set approximately 100 times that
          number of chunks.

        .. seealso::

            http://docs.h5py.org/en/stable/high/file.html#chunk-cache

        """

        # -------------------- Chunk & RDCC Vals ------------------------------
        schema_shape = self.schema_shape
        itemsize = np.dtype(self.schema_dtype).itemsize
        expectedrows = COLLECTION_SIZE * COLLECTION_COUNT
        maindim = 0

        chunk_shape = calc_chunkshape(schema_shape, expectedrows, itemsize, maindim)
        if chunk_shape == (1,) and schema_shape == ():
            schema_shape = (1,)
        req_chunks_per_dim = [math.ceil(i / j) for i, j in zip(schema_shape, chunk_shape)]
        req_shape = [i * j for i, j in zip(req_chunks_per_dim, chunk_shape)]
        chunk_nbytes = np.prod(chunk_shape) * itemsize
        nchunks = np.prod(req_chunks_per_dim)

        rdcc_nbytes_val = chunk_nbytes * nchunks * COLLECTION_SIZE
        if rdcc_nbytes_val >= CHUNK_MAX_RDCC_NBYTES:
            rdcc_nbytes_val = CHUNK_MAX_RDCC_NBYTES
        rdcc_nslots_guess = nchunks * expectedrows * 100
        rdcc_nslots_prime_val = find_next_prime(rdcc_nslots_guess)

        # ---------------------------- File Creation --------------------------

        uid = random_string()
        file_path = Path(self.DATADIR, f'{uid}.hdf5')
        self.wFp[uid] = h5py.File(file_path,
                                  mode='w',
                                  libver='latest',
                                  rdcc_nbytes=rdcc_nbytes_val,
                                  rdcc_w0=CHUNK_RDCC_W0,
                                  rdcc_nslots=rdcc_nslots_prime_val)
        self.w_uid = uid
        self.wdset = None
        self.hNextPath = 0
        self.hIdx = 0
        self.hColsRemain = COLLECTION_COUNT
        self.hMaxSize = COLLECTION_SIZE

        process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
        Path(process_dir, f'{uid}.hdf5').touch()

        # ----------------------- Dataset Creation ----------------------------

        optKwargs = self._dataset_opts(**self._dflt_backend_opts)
        for dset_num in range(COLLECTION_COUNT):
            self.wFp[uid].create_dataset(
                f'/{dset_num}',
                shape=(COLLECTION_SIZE, *req_shape),
                dtype=self.schema_dtype,
                chunks=(1, *chunk_shape),
                **optKwargs)

        # ---------------------- Attribute Config Vals ------------------------

        self.wFp[self.w_uid]['/'].attrs['HANGAR_VERSION'] = __version__
        self.wFp[self.w_uid]['/'].attrs['schema_shape'] = self.schema_shape
        self.wFp[self.w_uid]['/'].attrs['schema_dtype_num'] = np.dtype(self.schema_dtype).num
        self.wFp[self.w_uid]['/'].attrs['next_location'] = (0, 0)
        self.wFp[self.w_uid]['/'].attrs['collection_max_size'] = COLLECTION_SIZE
        self.wFp[self.w_uid]['/'].attrs['collection_total'] = COLLECTION_COUNT
        self.wFp[self.w_uid]['/'].attrs['collections_remaining'] = COLLECTION_COUNT
        self.wFp[self.w_uid]['/'].attrs['rdcc_nbytes'] = rdcc_nbytes_val
        self.wFp[self.w_uid]['/'].attrs['rdcc_w0'] = CHUNK_RDCC_W0
        self.wFp[self.w_uid]['/'].attrs['rdcc_nslots'] = rdcc_nslots_prime_val
        self.wFp[self.w_uid]['/'].attrs['chunk_shape'] = chunk_shape
        if optKwargs['compression_opts'] is not None:
            self.wFp[self.w_uid]['/'].attrs['compression_opts'] = optKwargs['compression_opts']
        else:
            self.wFp[self.w_uid]['/'].attrs['compression_opts'] = False

        self.wFp[self.w_uid].flush()
        try:
            self.wFp[self.w_uid].swmr_mode = True
        except ValueError:
            assert self.wFp[self.w_uid].swmr_mode is True
        self.wdset = self.wFp[self.w_uid][f'/{self.hNextPath}']

    def read_data(self, hashVal: HDF5_01_DataHashSpec) -> np.ndarray:
        """Read data from an hdf5 file handle at the specified locations

        Parameters
        ----------
        hashVal : HDF5_01_DataHashSpec
            record specification parsed from its serialized store val in lmdb.

        Returns
        -------
        np.array
            requested data
        """
        dsetCol = f'/{hashVal.dataset}'
        srcSlc = (hashVal.dataset_idx, *[slice(0, dim) for dim in hashVal.shape])
        rdictkey = f'{hashVal.uid}{dsetCol}'

        if self.schema_dtype:  # if is not None
            destArr = np.empty(hashVal.shape, self.schema_dtype)
            if rdictkey in self.rDatasets:
                self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
            else:
                try:
                    self.Fp[hashVal.uid][dsetCol].read_direct(destArr, srcSlc, None)
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except TypeError:
                    self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                    self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
                except KeyError:
                    process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
                    if Path(process_dir, f'{hashVal.uid}.hdf5').is_file():
                        file_pth = self.DATADIR.joinpath(f'{hashVal.uid}.hdf5')
                        self.rFp[hashVal.uid] = h5py.File(file_pth, 'r', swmr=True, libver='latest')
                        self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                        self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
                    else:
                        raise
        else:
            if rdictkey in self.rDatasets:
                destArr = self.rDatasets[rdictkey][srcSlc]
            else:
                try:
                    destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except TypeError:
                    self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
                    destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except KeyError:
                    process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
                    if Path(process_dir, f'{hashVal.uid}.hdf5').is_file():
                        file_pth = self.DATADIR.joinpath(f'{hashVal.uid}.hdf5')
                        self.rFp[hashVal.uid] = h5py.File(file_pth, 'r', swmr=True, libver='latest')
                        destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                        self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                    else:
                        raise

        if xxh64_hexdigest(destArr) != hashVal.checksum:
            # try casting to check if dtype does not match for all zeros case
            destArr = destArr.astype(np.typeDict[self.Fp[hashVal.uid]['/'].attrs['schema_dtype_num']])
            if xxh64_hexdigest(destArr) != hashVal.checksum:
                raise RuntimeError(
                    f'DATA CORRUPTION Checksum {xxh64_hexdigest(destArr)} != recorded {hashVal}')
        return destArr

    def write_data(self, array: np.ndarray, *, remote_operation: bool = False) -> bytes:
        """verifies correctness of array data and performs write operation.

        Parameters
        ----------
        array : np.ndarray
            tensor to write to group.
        remote_operation : optional, kwarg only, bool
            If this is a remote process which is adding data, any necessary
            hdf5 dataset files will be created in the remote data dir instead
            of the stage directory. (default is False, which is for a regular
            access process)

        Returns
        -------
        bytes
            string identifying the collection dataset and collection dim-0 index
            which the array can be accessed at.
        """
        checksum = xxh64_hexdigest(array)
        if self.w_uid in self.wFp:
            self.hIdx += 1
            if self.hIdx >= self.hMaxSize:
                self.wdset.flush()
                self.hIdx = 0
                self.hNextPath += 1
                self.hColsRemain -= 1
                self.wdset = self.wFp[self.w_uid][f'/{self.hNextPath}']
                if self.hColsRemain <= 1:
                    self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
                    self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
                    self.wFp[self.w_uid].flush()
                    self._create_schema(remote_operation=remote_operation)
        else:
            self._create_schema(remote_operation=remote_operation)

        destSlc = (self.hIdx, *[slice(0, dim) for dim in array.shape])
        self.wdset.write_direct(array, None, destSlc)
        self.wdset.flush()
        res = hdf5_01_encode(self.w_uid, checksum, self.hNextPath, self.hIdx, array.shape)
        return res



"""Accessor column containing nested mapping of data under top level keys.
"""
from contextlib import ExitStack
from pathlib import Path
from typing import (
    Tuple, Union, Dict, Iterable, Any, Optional
)
from weakref import proxy

from .common import open_file_handles
from ..records import (
    data_record_db_val_from_digest,
    data_record_digest_val_from_db_val,
    nested_data_db_key_from_names,
    hash_data_db_key_from_raw_key,
    schema_db_key_from_column,
    schema_hash_db_key_from_digest,
    schema_hash_record_db_val_from_spec,
    schema_record_db_val_from_digest,
)
from ..records.parsing import generate_sample_name
from ..backends import backend_decoder, BACKEND_ACCESSOR_MAP
from ..op_state import reader_checkout_only
from ..utils import is_suitable_user_key, valfilter, valfilterfalse


KeyType = Union[str, int]
EllipsisType = type(Ellipsis)
GetKeysType = Union[KeyType, EllipsisType, slice]


class FlatSubsampleReader(object):

    __slots__ = ('_column_name', '_stack', '_be_fs',
                 '_mode', '_subsamples', '_samplen')
    _attrs = __slots__

    def __init__(self,
                 columnname: str,
                 samplen: str,
                 be_handles: BACKEND_ACCESSOR_MAP,
                 specs,
                 mode: str,
                 *args, **kwargs):

        self._column_name = columnname
        self._samplen = samplen
        self._be_fs = be_handles
        self._subsamples = specs
        self._mode = mode
        self._stack: Optional[ExitStack] = None

    @property
    def _debug_(self):  # pragma: no cover
        return {
            '__class__': self.__class__,
            '_column_name': self._column_name,
            '_samplen': self._samplen,
            '_be_fs': self._be_fs,
            '_subsamples': self._subsamples,
            '_mode': self._mode,
            '_stack': self._stack._exit_callbacks if self._stack else self._stack,
        }

    def __repr__(self):
        res = f'{self.__class__}('\
              f'column_name={self._column_name}, '\
              f'sample_name={self._samplen})'
        return res

    def _repr_pretty_(self, p, cycle):
        res = f'Hangar {self.__class__.__name__} \
                \n    Column Name          : {self._column_name}\
                \n    Sample Name          : {self._samplen}\
                \n    Writeable            : "{self.iswriteable}"\
                \n    Number of Subsamples : {len(self)}\n'
        p.text(res)

    def _ipython_key_completions_(self):
        """Let ipython know that any key based access can use the column keys

        Since we don't want to inherit from dict, nor mess with `__dir__` for
        the sanity of developers, this is the best way to ensure users can
        autocomplete keys.

        Returns
        -------
        list
            list of strings, each being one of the column keys for access.
        """
        return list(self.keys())

    def __enter__(self):
        self._enter_count += 1
        return self

    def __exit__(self, *exc):
        self._enter_count -= 1

    def _destruct(self):
        if isinstance(self._stack, ExitStack):
            self._stack.close()
        for attr in self._attrs:
            delattr(self, attr)

    def __getattr__(self, name):
        """Raise permission error after checkout is closed.

         Only runs after a call to :meth:`_destruct`, which is responsible for
         deleting all attributes from the object instance.
        """
        try:
            self.__getattribute__('_mode')  # once checkout is closed, this won't exist.
        except AttributeError:
            err = (f'Unable to operate on past checkout objects which have been '
                   f'closed. No operation occurred. Please use a new checkout.')
            raise PermissionError(err) from None
        return self.__getattribute__(name)

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        return {slot: getattr(self, slot) for slot in self.__slots__}

    def __setstate__(self, state: dict) -> None:
        """ensure multiprocess operations can pickle relevant data.

        Technically should be decorated with @reader_checkout_only, but since
        at instance creation that is not an attribute, the decorator won't
        know. Since only readers can be pickled, This isn't much of an issue.
        """
        for slot, value in state.items():
            setattr(self, slot, value)

    def __len__(self) -> int:
        return len(self._subsamples)

    def __contains__(self, key: KeyType) -> bool:
        return key in self._subsamples

    def __iter__(self) -> Iterable[KeyType]:
        yield from self.keys()

    def __getitem__(self, key: GetKeysType) -> Union[Any, Dict[KeyType, Any]]:
        """Retrieve data for some subsample key via dict style access conventions.

        .. seealso:: :meth:`get`

        Parameters
        ----------
        key : GetKeysType
            Sample key to retrieve from the column. Alternatively, ``slice``
            syntax can be used to retrieve a selection of subsample
            keys/values. An empty slice (``: == slice(None)``) or ``Ellipsis``
            (``...``) will return all subsample keys/values. Passing a
            non-empty slice (``[1:5] == slice(1, 5)``) will select keys to
            retrieve by enumerating all subsamples and retrieving the element
            (key) for each step across the range. Note: order of enumeration is
            not guaranteed; do not rely on any ordering observed when using
            this method.

        Returns
        -------
        Union[Any, Dict[KeyType, Any]]
            Sample data corresponding to the provided key. or dictionary
            of subsample keys/data if Ellipsis or slice passed in as key.

        Raises
        ------
        KeyError
            if no sample with the requested key exists.
        """
        # select subsample(s) with regular keys
        if isinstance(key, (str, int)):
            spec = self._subsamples[key]
            return self._be_fs[spec.backend].read_data(spec)
        # select all subsamples
        elif key is Ellipsis:
            res = {}
            for subsample, spec in self._subsamples.items():
                res[subsample] = self._be_fs[spec.backend].read_data(spec)
            return res
        # slice subsamples by sorted order of keys
        elif isinstance(key, slice):
            res = {}
            subsample_spec_slice = tuple(self._subsamples.items())[key]
            for subsample, spec in subsample_spec_slice:
                spec = self._subsamples[subsample]
                res[subsample] = self._be_fs[spec.backend].read_data(spec)
            return res
        else:
            raise TypeError(f'key {key} type {type(key)} not valid.')

    @property
    def _enter_count(self):
        return self._be_fs['enter_count']

    @_enter_count.setter
    def _enter_count(self, value):
        self._be_fs['enter_count'] = value

    @property
    def _is_conman(self):
        return bool(self._enter_count)

    @property
    def sample(self) -> KeyType:
        """Name of the sample this column subsamples are stured under.
        """
        return self._samplen

    @property
    def column(self) -> str:
        """Name of the column.
        """
        return self._column_name

    @property
    def iswriteable(self) -> bool:
        """Bool indicating if this column object is write-enabled.
        """
        return False if self._mode == 'r' else True

    @property
    def data(self) -> Dict[KeyType, Any]:
        """Return dict mapping every subsample key / data value stored in the sample.

        Returns
        -------
        Dict[KeyType, Any]
            Dictionary mapping subsample name(s) (keys) to their stored values
            as :class:`numpy.ndarray` instances.
        """
        return self[...]

    def _mode_local_aware_key_looper(self, local: bool) -> Iterable[KeyType]:
        """Generate keys for iteration with dict update safety ensured.

        Parameters
        ----------
        local : bool
            True if keys should be returned which only exist on the local
            machine. False if remote sample keys should be excluded.

        Returns
        -------
        Iterable[KeyType]
            Sample keys conforming to the `local` argument spec.
        """
        if local:
            if self._mode == 'r':
                yield from valfilter(lambda x: x.islocal, self._subsamples).keys()
            else:
                yield from tuple(valfilter(lambda x: x.islocal, self._subsamples).keys())
        else:
            if self._mode == 'r':
                yield from self._subsamples.keys()
            else:
                yield from tuple(self._subsamples.keys())

    @property
    def contains_remote_references(self) -> bool:
        """Bool indicating all subsamples in sample column exist on local disk.

        The data associated with subsamples referencing some remote server will
        need to be downloaded (``fetched`` in the hangar vocabulary) before
        they can be read into memory.

        Returns
        -------
        bool
            False if at least one subsample in the column references data
            stored on some remote server. True if all sample data is available
            on the machine's local disk.
        """
        return not all(map(lambda x: x.islocal, self._subsamples.values()))

    @property
    def remote_reference_keys(self) -> Tuple[KeyType]:
        """Compute subsample names whose data is stored in a remote server reference.

        Returns
        -------
        Tuple[KeyType]
            list of subsample keys in the column whose data references indicate
            they are stored on a remote server.
        """
        return tuple(valfilterfalse(lambda x: x.islocal, self._subsamples).keys())

    def keys(self, local: bool = False) -> Iterable[KeyType]:
        """Generator yielding the name (key) of every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys will only correspond to data which is
            available for reading on the local disk, by default False.

        Yields
        ------
        Iterable[KeyType]
            Keys of one subsample at a time inside the sample.
        """
        yield from self._mode_local_aware_key_looper(local)

    def values(self, local: bool = False) -> Iterable[Any]:
        """Generator yielding the data for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned values will only correspond to data which is
            available for reading on the local disk. No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Any]
            Values of one subsample at a time inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield self[key]

    def items(self, local: bool = False) -> Iterable[Tuple[KeyType, Any]]:
        """Generator yielding (name, data) tuple for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys/values will only correspond to data which is
            available for reading on the local disk, No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Tuple[KeyType, Any]]
            Name and stored value for every subsample inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield (key, self[key])

    def get(self, key: KeyType, default=None):
        """Retrieve the data associated with some subsample key

        Parameters
        ----------
        key : GetKeysType
            The name of the subsample(s) to retrieve. Passing a single
            subsample key will return the stored :class:`numpy.ndarray`
        default
            if a `key` parameter is not found, then return this value instead.
            By default, None.

        Returns
        -------
        value
            data stored under subsample key if key exists, else default
            value if not found.
        """
        try:
            return self[key]
        except KeyError:
            return default


# ---------------- writer methods only after this point -------------------


class FlatSubsampleWriter(FlatSubsampleReader):

    __slots__ = ('_schema', '_txnctx', '_path')
    _attrs = __slots__ + FlatSubsampleReader.__slots__

    def __init__(self,
                 schema,
                 repo_path: Path,
                 aset_ctx=None,
                 *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._path = repo_path
        self._schema = schema
        self._txnctx = aset_ctx

    def __enter__(self):
        with ExitStack() as stack:
            self._txnctx.open_write()
            stack.callback(self._txnctx.close_write)
            if self._enter_count == 0:
                for k in self._be_fs.keys():
                    if k in ('enter_count', 'schema_spec'):
                        continue
                    stack.enter_context(self._be_fs[k])
            self._enter_count += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._stack.close()
        self._enter_count -= 1
        if self._enter_count == 0:
            self._stack = None

    def _set_arg_validate(self, key, value):
        if not is_suitable_user_key(key):
            raise ValueError(f'Sample name `{key}` is not suitable.')
        isCompat = self._schema.verify_data_compatible(value)
        if not isCompat.compatible:
            raise ValueError(isCompat.reason)

    def _perform_set(self, key, value):
        """Internal write method. Assumes all arguments validated and cm open.

        Parameters
        ----------
        key
            subsample key to store
        value
            data to store
        """
        # full_hash = ndarray_hasher_tcode_0(value)
        full_hash = self._schema.data_hash_digest(value)
        hashKey = hash_data_db_key_from_raw_key(full_hash)

        # check if data record already exists with given key
        dataRecKey = nested_data_db_key_from_names(self._column_name, self._samplen, key)
        existingDataRecVal = self._txnctx.dataTxn.get(dataRecKey, default=False)
        if existingDataRecVal:
            # check if data record already with same key & hash value
            existingDataRec = data_record_digest_val_from_db_val(existingDataRecVal)
            if full_hash == existingDataRec.digest:
                return

        # write new data if data hash does not exist
        existingHashVal = self._txnctx.hashTxn.get(hashKey, default=False)
        if existingHashVal is False:
            backendCode = self._schema.backend
            hashVal = self._be_fs[backendCode].write_data(value)
            self._txnctx.hashTxn.put(hashKey, hashVal)
            self._txnctx.stageHashTxn.put(hashKey, hashVal)
            hash_spec = backend_decoder(hashVal)
        else:
            hash_spec = backend_decoder(existingHashVal)
            if hash_spec.backend not in self._be_fs:
                # when adding data which is already stored in the repository, the
                # backing store for the existing data location spec may not be the
                # same as the backend which the data piece would have been saved in here.
                #
                # As only the backends actually referenced by a columns samples are
                # initialized (accessible by the column), there is no guarantee that
                # an accessor exists for such a sample. In order to prevent internal
                # errors from occurring due to an uninitialized backend if a previously
                # existing data piece is "saved" here and subsequently read back from
                # the same writer checkout, we perform an existence check and backend
                # initialization, if appropriate.
                fh = open_file_handles(backends=(hash_spec.backend,),
                                       path=self._path,
                                       mode='a',
                                       schema=self._schema)
                self._be_fs[hash_spec.backend] = fh[hash_spec.backend]

        # add the record to the db
        dataRecVal = data_record_db_val_from_digest(full_hash)
        self._txnctx.dataTxn.put(dataRecKey, dataRecVal)
        self._subsamples[key] = hash_spec

    def __setitem__(self, key, value):
        """Store data as a subsample. Convenience method to :meth:`add`.

        .. seealso::

            :meth:`update` for an implementation analogous to python's built
            in :meth:`dict.update` method which accepts a dict or iterable of
            key/value pairs to add in the same operation.

        Parameters
        ----------
        key
            Key (name) of the subsample to add to the column.
        value
            Data to add as the sample.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)
            self._set_arg_validate(key, value)
            self._perform_set(key, value)

    def append(self, value) -> KeyType:
        """Store some data in a subsample with an automatically generated key.

        This method should only be used if the context some piece of data is
        used in is independent from it's value (ie. when reading data back,
        there is no useful information which needs to be conveyed between the
        data source's name/id and the value of that piece of information.)
        Think carefully before going this route, as this posit does not apply
        to many common use cases.

        .. seealso::

            In order to store the data with a user defined key, use
            :meth:`update` or :meth:`__setitem__`

        Parameters
        ----------
        value
            Piece of data to store in the column.

        Returns
        -------
        KeyType
            Name of the generated key this data is stored with.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)
            key = generate_sample_name()
            while key in self._subsamples:
                key = generate_sample_name()
            self._set_arg_validate(key, value)
            self._perform_set(key, value)
            return key

    def update(self, other=None, **kwargs):
        """Store data with the key/value pairs, overwriting existing keys.

        :meth:`update` implements functionality similar to python's builtin
        :meth:`dict.update` method, accepting either a dictionary or other
        iterable (of length two) listing key / value pairs.

        Parameters
        ----------
        other
            Accepts either another dictionary object or an iterable of
            key/value pairs (as tuples or other iterables of length two).
            mapping sample names to data values, If sample name is string type,
            can only contain alpha-numeric ascii characters (in addition to
            '-', '.', '_'). Int key must be >= 0. By default, None.
        **kwargs
            keyword arguments provided will be saved with keywords as subsample
            keys (string type only) and values as np.array instances.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if other:
                if not isinstance(other, dict):
                    other = dict(other)
                else:
                    other = other.copy()
            elif other is None:
                other = {}
            if kwargs:
                # we have to merge kwargs dict with `other` before operating on
                # either so all validation and writing occur atomically
                other.update(kwargs)

            for key, val in other.items():
                self._set_arg_validate(key, val)
            for key, val in other.items():
                self._perform_set(key, val)

    def __delitem__(self, key: KeyType):
        """Remove a subsample from the column.`.

        .. seealso::

            :meth:`pop` to simultaneously get a keys value and delete it.

        Parameters
        ----------
        key : KeyType
            Name of the sample to remove from the column.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if key not in self._subsamples:
                raise KeyError(key)

            dbKey = nested_data_db_key_from_names(self._column_name, self._samplen, key)
            isRecordDeleted = self._txnctx.dataTxn.delete(dbKey)
            if isRecordDeleted is False:
                raise RuntimeError(
                    f'Internal error. Not able to delete key {key} from staging '
                    f'db even though existence passed in memory verification. '
                    f'Please report this message in full to the hangar development team.',
                    f'Specified key: <{type(key)} {key}>', f'Calculated dbKey: <{dbKey}>',
                    f'isRecordDeleted: <{isRecordDeleted}>', f'DEBUG STRING: {self._debug_}')
            del self._subsamples[key]

    def pop(self, key: KeyType):
        """Retrieve some value for some key(s) and delete it in the same operation.

        Parameters
        ----------
        key : KeysType
            Sample key to remove

        Returns
        -------
        value
            Upon success, the value of the removed key.
        """
        value = self[key]
        del self[key]
        return value


class NestedSampleReader:

    __slots__ = ('_mode', '_column_name', '_samples',
                 '_be_fs', '_path', '_stack', '_schema')
    _attrs = __slots__

    def __init__(self,
                 columnname: str,
                 samples: Dict[KeyType, FlatSubsampleReader],
                 backend_handles: Dict[str, Any],
                 repo_path: Path,
                 mode: str,
                 schema=None,
                 *args, **kwargs):

        self._mode = mode
        self._column_name = columnname
        self._samples = samples
        self._be_fs = backend_handles
        self._path = repo_path
        self._stack: Optional[ExitStack] = None
        self._schema = schema

    def __repr__(self):
        res = (
            f'{self.__class__.__qualname__}('
            f'repo_pth={self._path}, '
            f'columnname={self._column_name}, '
            f"{[f'{key}={val}, ' for key, val in self._schema.schema.items()]}, "
            f'mode={self._mode})')
        return res

    def _repr_pretty_(self, p, cycle):
        res = f'Hangar {self.__class__.__qualname__} \
                \n    Column Name              : {self.column}\
                \n    Writeable                : {self.iswriteable}\
                \n    Column Type              : {self.column_type}\
                \n    Column Layout            : {self.column_layout}\
                \n    Schema Type              : {self.schema_type}\
                \n    DType                    : {self.dtype}\
                \n    Shape                    : {self.shape}\
                \n    Number of Samples        : {len(self)}\
                \n    Number of Subsamples     : {self.num_subsamples}\
                \n    Partial Remote Data Refs : {bool(self.contains_remote_references)}\n'
        p.text(res)

    def _ipython_key_completions_(self):
        """Let ipython know that any key based access can use the column keys

        Since we don't want to inherit from dict, nor mess with `__dir__` for
        the sanity of developers, this is the best way to ensure users can
        autocomplete keys.

        Returns
        -------
        list
            list of strings, each being one of the column keys for access.
        """
        return list(self.keys())

    def __enter__(self):
        self._enter_count += 1
        return self

    def __exit__(self, *exc):
        self._enter_count -= 1

    def _destruct(self):
        if isinstance(self._stack, ExitStack):
            self._stack.close()
        self._close()
        for sample in self._samples.values():
            sample._destruct()
        for attr in self._attrs:
            delattr(self, attr)

    def __getattr__(self, name):
        """Raise permission error after checkout is closed.

         Only runs after a call to :meth:`_destruct`, which is responsible
         for deleting all attributes from the object instance.
        """
        try:
            self.__getattribute__('_mode')  # once checkout is closed, this won't exist.
        except AttributeError:
            err = (f'Unable to operate on past checkout objects which have been '
                   f'closed. No operation occurred. Please use a new checkout.')
            raise PermissionError(err) from None
        return self.__getattribute__(name)

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        return {slot: getattr(self, slot) for slot in self.__slots__}

    def __setstate__(self, state: dict) -> None:
        """ensure multiprocess operations can pickle relevant data.

        Technically should be decorated with @reader_checkout_only, but since
        at instance creation the '_mode' is not a set attribute, the decorator
        won't know how to process. Since only readers can be pickled, This
        isn't much of an issue.
        """
        for slot, value in state.items():
            setattr(self, slot, value)

    def __getitem__(self, key: KeyType) -> FlatSubsampleReader:
        """Get the sample access class for some sample key.

        Parameters
        ----------
        key : KeyType
            Name of sample to retrieve

        Returns
        -------
        FlatSubsampleReader
            Sample accessor corresponding to the given key

        Raises
        ------
        KeyError
            If no sample with the provided key exists.
        """
        return self._samples[key]

    def __iter__(self) -> Iterable[KeyType]:
        """Create iterator yielding an column sample keys.

        Yields
        -------
        Iterable[KeyType]
            Sample key contained in the column.
        """
        yield from self.keys()

    def __len__(self) -> int:
        """Find number of samples in the column
        """
        return len(self._samples)

    def __contains__(self, key: KeyType) -> bool:
        """Determine if some sample key exists in the column.
        """
        return key in self._samples

    def _open(self):
        for val in self._be_fs.values():
            try:
                # since we are storing non backend accessor information in the
                # be_fs weakref proxy for the purpose of memory savings, not
                # all elements have a `open` method
                val.open(mode=self._mode)
            except AttributeError:
                pass

    def _close(self):
        for val in self._be_fs.values():
            # since we are storing non backend accessor information in the
            # be_fs weakref proxy for the purpose of memory savings, not all
            # elements have a `close` method
            try:
                val.close()
            except AttributeError:
                pass

    @property
    def _enter_count(self):
        return self._be_fs['enter_count']

    @_enter_count.setter
    def _enter_count(self, value):
        self._be_fs['enter_count'] = value

    @property
    def _is_conman(self):
        return bool(self._enter_count)

    @property
    def column(self) -> str:
        """Name of the column.
        """
        return self._column_name

    @property
    def column_type(self):
        """Data container type of the column ('ndarray', 'str', etc).
        """
        return self._schema.column_type

    @property
    def column_layout(self):
        """Column layout type ('nested', 'flat', etc).
        """
        return self._schema.column_layout

    @property
    def schema_type(self):
        """Schema type of the contained data ('variable_shape', 'fixed_shape', etc).
        """
        return self._schema.schema_type

    @property
    def dtype(self):
        """Dtype of the columns data (np.float, str, etc).
        """
        return self._schema.dtype

    @property
    def shape(self):
        """(Max) shape of data that can (is) written in the column.
        """
        try:
            return self._schema.shape
        except AttributeError:
            return None

    @property
    def backend(self) -> str:
        """Code indicating which backing store is used when writing data.
        """
        return self._schema.backend

    @property
    def backend_options(self):
        """Filter / Compression options applied to backend when writing data.
        """
        return self._schema.backend_options

    @property
    def iswriteable(self) -> bool:
        """Bool indicating if this column object is write-enabled.
        """
        return False if self._mode == 'r' else True

    def _mode_local_aware_key_looper(self, local: bool) -> Iterable[KeyType]:
        """Generate keys for iteration with dict update safety ensured.

        Parameters
        ----------
        local : bool
            True if keys should be returned which only exist on the local
            machine. False if remote sample keys should be excluded.

        Returns
        -------
        Iterable[KeyType]
            Sample keys conforming to the `local` argument spec.
        """
        if local:
            if self._mode == 'r':
                yield from valfilterfalse(lambda x: x.contains_remote_references, self._samples).keys()
            else:
                yield from tuple(valfilterfalse(lambda x: x.contains_remote_references, self._samples).keys())
        else:
            if self._mode == 'r':
                yield from self._samples.keys()
            else:
                yield from tuple(self._samples.keys())

    @property
    def contains_remote_references(self) -> bool:
        """Bool indicating all subsamples in sample column exist on local disk.

        The data associated with subsamples referencing some remote server will
        need to be downloaded (``fetched`` in the hangar vocabulary) before
        they can be read into memory.

        Returns
        -------
        bool
            False if at least one subsample in the column references data
            stored on some remote server. True if all sample data is available
            on the machine's local disk.
        """
        return all(map(lambda x: x.contains_remote_references, self._samples.values()))

    @property
    def remote_reference_keys(self) -> Tuple[KeyType]:
        """Compute subsample names whose data is stored in a remote server reference.

        Returns
        -------
        Tuple[KeyType]
            list of subsample keys in the column whose data references indicate
            they are stored on a remote server.
        """
        return tuple(valfilter(lambda x: x.contains_remote_references, self._samples).keys())

    @property
    def contains_subsamples(self) -> bool:
        """Bool indicating if sub-samples are contained in this column container.
        """
        return True

    @property
    def num_subsamples(self) -> int:
        """Calculate total number of subsamples existing in all samples in column
        """
        total = 0
        for sample in self._samples.values():
            total += len(sample)
        return total

    def keys(self, local: bool = False) -> Iterable[KeyType]:
        """Generator yielding the name (key) of every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys will only correspond to data which is
            available for reading on the local disk, by default False.

        Yields
        ------
        Iterable[KeyType]
            Keys of one subsample at a time inside the sample.
        """
        yield from self._mode_local_aware_key_looper(local)

    def values(self, local: bool = False) -> Iterable[Any]:
        """Generator yielding the tensor data for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned values will only correspond to data which is
            available for reading on the local disk. No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Any]
            Values of one subsample at a time inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield self[key]

    def items(self, local: bool = False) -> Iterable[Tuple[KeyType, Any]]:
        """Generator yielding (name, data) tuple for every subsample.

        Parameters
        ----------
        local : bool, optional
            If True, returned keys/values will only correspond to data which is
            available for reading on the local disk, No attempt will be made to
            read data existing on a remote server, by default False.

        Yields
        ------
        Iterable[Tuple[KeyType, Any]]
            Name and stored value for every subsample inside the sample.
        """
        for key in self._mode_local_aware_key_looper(local):
            yield (key, self[key])

    def get(self, key: GetKeysType, default: Any = None) -> FlatSubsampleReader:
        """Retrieve data for some sample key(s) in the column.

        Parameters
        ----------
        key : GetKeysType
            The name of the subsample(s) to retrieve
        default : Any
            if a `key` parameter is not found, then return this value instead.
            By default, None.

        Returns
        -------
        FlatSubsampleReader:
            Sample accessor class given by name ``key`` which can be used to
            access subsample data.
        """
        try:
            return self[key]
        except KeyError:
            return default


# ---------------- writer methods only after this point -------------------


class NestedSampleWriter(NestedSampleReader):

    __slots__ = ('_txnctx',)
    _attrs = __slots__ + NestedSampleReader.__slots__

    def __init__(self, aset_ctx=None, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._txnctx = aset_ctx

    def __enter__(self):
        with ExitStack() as stack:
            self._txnctx.open_write()
            stack.callback(self._txnctx.close_write)
            if self._enter_count == 0:
                for k in tuple(self._be_fs.keys()):
                    if k in ('enter_count', 'schema_spec'):
                        continue
                    stack.enter_context(self._be_fs[k])
            self._enter_count += 1
            self._stack = stack.pop_all()
        return self

    def __exit__(self, *exc):
        self._stack.close()
        self._enter_count -= 1

    def _set_arg_validate(self, sample_key, subsample_map):
        if not is_suitable_user_key(sample_key):
            raise ValueError(f'Sample name `{sample_key}` is not suitable.')

        for subsample_key, subsample_val in subsample_map.items():
            if not is_suitable_user_key(subsample_key):
                raise ValueError(f'Sample name `{sample_key}` is not suitable.')
            isCompat = self._schema.verify_data_compatible(subsample_val)
            if not isCompat.compatible:
                raise ValueError(isCompat.reason)

    def _perform_set(self, key, value) -> None:
        if key in self._samples:
            self._samples[key].update(value)
        else:
            self._samples[key] = FlatSubsampleWriter(
                schema=proxy(self._schema),
                aset_ctx=proxy(self._txnctx),
                repo_path=self._path,
                columnname=self._column_name,
                samplen=key,
                be_handles=proxy(self._be_fs),
                specs={},
                mode='a')
            try:
                self._samples[key].update(value)
            except Exception as e:
                del self._samples[key]
                raise e

    def __setitem__(self, key, value) -> None:
        """Store some subsample key / subsample data map, overwriting existing keys.

        .. seealso::

            :meth:`update` for alternative syntax for setting values.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)
            value = dict(value)
            self._set_arg_validate(key, value)
            self._perform_set(key, value)

    def update(self, other=None, **kwargs) -> None:
        """Store some data with the key/value pairs, overwriting existing keys.

        :meth:`update` implements functionality similar to python's builtin
        :meth:`dict.update` method, accepting either a dictionary or other
        iterable (of length two) listing key / value pairs.

        Parameters
        ----------
        other
            Dictionary mapping sample names to subsample data maps. Or Sequence
            (list or tuple) where element one is the sample name and element
            two is a subsample data map.
        **kwargs
            keyword arguments provided will be saved with keywords as sample
            keys (string type only) and values as a mapping of subarray keys
            to data values.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if isinstance(other, dict):
                other = other.copy()
            elif other:
                other = dict(other)
            else:
                other = {}
            if kwargs:
                # we merge kwargs dict with `other` before operating on either
                # so all necessary validation and writing occur atomically
                other.update(kwargs)
            for sample in tuple(other.keys()):
                other[sample] = dict(other[sample])

            for key, val in other.items():
                self._set_arg_validate(key, val)
            for key, val in other.items():
                self._perform_set(key, val)

    def __delitem__(self, key: KeyType):
        """Remove a sample (including all contained subsamples) from the column.

        .. seealso::

            :meth:`pop` for alternative implementing a simultaneous get value
            and delete operation.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            sample = self._samples[key]
            subsample_keys = list(sample.keys())
            for subkey in subsample_keys:
                del sample[subkey]

            self._samples[key]._destruct()
            del self._samples[key]

    def pop(self, key: KeyType) -> Dict[KeyType, Any]:
        """Retrieve some value for some key(s) and delete it in the same operation.

        Parameters
        ----------
        key : KeysType
            sample key to remove

        Returns
        -------
        Dict[KeyType, KeyArrMap]
            Upon success, a nested dictionary mapping sample names to a dict of
            subsample names and subsample values for every sample key passed
            into this method.
        """
        res = self._samples[key].data
        del self[key]
        return res

    def change_backend(self, backend: str, backend_options: Optional[dict] = None):
        """Change the default backend and filters applied to future data writes.

        .. warning::

           This method is meant for advanced users only. Please refer to the
           hangar backend codebase for information on accepted parameters and
           options.

        Parameters
        ----------
        backend : str
            Backend format code to swtich to.
        backend_options
            Backend option specification to use (if specified). If left to
            default value of None, then default options for backend are
            automatically used.

        Raises
        ------
        RuntimeError
            If this method was called while this column is invoked in a
            context manager
        ValueError
            If the backend format code is not valid.
        """
        if self._is_conman:
            raise RuntimeError('Cannot call method inside column context manager.')

        self._schema.change_backend(backend, backend_options=backend_options)

        new_schema_digest = self._schema.schema_hash_digest()
        columnSchemaKey = schema_db_key_from_column(self._column_name, layout=self.column_layout)
        columnSchemaVal = schema_record_db_val_from_digest(new_schema_digest)
        hashSchemaKey = schema_hash_db_key_from_digest(new_schema_digest)
        hashSchemaVal = schema_hash_record_db_val_from_spec(self._schema.schema)

        # -------- set vals in lmdb only after schema is sure to exist --------

        with self._txnctx.write() as ctx:
            ctx.dataTxn.put(columnSchemaKey, columnSchemaVal)
            ctx.hashTxn.put(hashSchemaKey, hashSchemaVal, overwrite=False)

        new_backend = self._schema.backend
        if new_backend not in self._be_fs:
            fhands = open_file_handles(
                backends=[new_backend],
                path=self._path,
                mode='a',
                schema=self._schema)
            self._be_fs[new_backend] = fhands[new_backend]
        else:
            self._be_fs[new_backend].close()
        self._be_fs[new_backend].open(mode='a')
        self._be_fs[new_backend].backend_opts = self._schema.backend_options
        return
